package fr.labri.collections;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import fr.labri.collections.ITree;
import fr.labri.collections.TreeUtils;
import fr.labri.collections.ioutils.IndentingXMLStreamWriter;

public final class TreeIoUtils {
	
	private final static QName TYPE = new QName("type");
	private final static QName LABEL = new QName("label");

	private TreeIoUtils() {
	}
	
	public static ITree fromXml(InputStream iStream) {
		return fromXml(new InputStreamReader(iStream));
	}
	
	public static ITree fromXmlString(String xml) {
		return fromXml(new StringReader(xml));
	}
	
	public static ITree fromXmlFile(String path) {
		try {
			return fromXml(new FileReader(path));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public static ITree fromXml(Reader source) {
		XMLInputFactory fact = XMLInputFactory.newInstance();
		ITree root = null;
		try {
			Stack<ITree> trees = new Stack<>();
			XMLEventReader r = fact.createXMLEventReader(source);
			while (r.hasNext()) {
				XMLEvent e = r.nextEvent();
				if (e instanceof StartElement) {
					StartElement s = (StartElement) e;
					int type = Integer.parseInt(s.getAttributeByName(TYPE).getValue());
					
					ITree t = new Tree(type, labelForAttribute(s, LABEL));
					
					if (trees.isEmpty())
						root = t;
					else
						t.setParentAndUpdateChildren(trees.peek());
					trees.push(t);
				} else if (e instanceof EndElement)
					trees.pop();
			}
			return root;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public static void toXml(ITree tree, String file) throws IOException {
		FileWriter f = new FileWriter(file);
		toXml(tree, f);
		f.close();
	}
	
	public static void toXml(ITree tree, Writer writer) {
		try {
			XMLOutputFactory f = XMLOutputFactory.newInstance();
			XMLStreamWriter w = new IndentingXMLStreamWriter(f.createXMLStreamWriter(writer));
			AbsXMLSerializer serializer = new XMLSerializer(w, tree);

			serialize(tree, serializer);
			w.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static String toXml(ITree ctx) {
		StringWriter s = new StringWriter();
		try {
			toXml(ctx, s);
			return s.toString();
		} finally {
			try {
				s.close();
			} catch (IOException e) { }
		}
	}
	
	public static String toDot(ITree root) { // FIXME should be a Serializer
		StringBuffer b = new StringBuffer();
		List<ITree> trees = TreeUtils.breadthFirst(root);
		
		Map<ITree, Integer> ids = new HashMap<>();
		b.append("digraph G {\n");
		for (int i = 0; i < trees.size(); i ++) {
			ITree t = trees.get(i);
			ids.put(t, i);
			String label = String.format("%d:%s", t.getType(), t.getLabel());
			if (label.contains("\"") || label.contains("\\s"))
				label = label.replaceAll("\"", "").replaceAll("\\s", "").replaceAll("\\\\", "");
			if (label.length() > 30)
				label = label.substring(0, 30);
			b.append(String.format("\t%d [label=\"%s\"];\n", i, label));
		}

		b.append("\n");
		
		for (int i = 0; i < trees.size(); i ++) {
			ITree t = trees.get(i);
			ITree p = t.getParent();
			if (ids.containsKey(p))
				b.append(String.format("\t%d -> %d;\n", ids.get(p), i));
		}
		b.append("}");
		return b.toString();
	}
	
	public static void toDot(ITree tree, String path) {
		try {
			FileWriter fw = new FileWriter(path);
			fw.append(toDot(tree));
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	static void serialize(ITree root, TreeSerializer serializer) throws Exception {
		serializer.startSerialization();
		writeTree(serializer, root);
		serializer.stopSerialization();
	}
	
	private static void writeTree(TreeSerializer s, ITree t) throws Exception {
		s.startTree(t);
		for (ITree c: t.getChildren())
			writeTree(s, c);
		s.endTree(t);
	}

	static String labelForAttribute(StartElement s, QName attrName) {
		Attribute attr = s.getAttributeByName(attrName);
		return attr == null ? ITree.NO_LABEL : attr.getValue();
	}

	static int numberForAttribute(StartElement s, QName attrName) {
		return Integer.parseInt(s.getAttributeByName(attrName).getValue());
	}
	
	interface TreeSerializer {
		void startSerialization() throws Exception;
		void startTree(ITree tree) throws Exception;
		void endTree(ITree tree) throws Exception;
		void stopSerialization() throws Exception;
	}
	
	static abstract class AbsXMLSerializer implements TreeSerializer {
		protected XMLStreamWriter writer;
		protected ITree root;

		protected AbsXMLSerializer(XMLStreamWriter w, ITree r) {
			writer = w;
			root = r;
		}

		@Override
		public void startSerialization() throws XMLStreamException {
			writer.writeStartDocument();
		}

		@Override
		public void stopSerialization() throws XMLStreamException {
			writer.writeEndDocument();
		}
	}
	
	static class XMLSerializer extends AbsXMLSerializer {
		public XMLSerializer(XMLStreamWriter w, ITree root) {
			super(w, root);
		}

		@Override
		public void startTree(ITree tree) throws XMLStreamException {
			writer.writeStartElement("tree");
			writer.writeAttribute("type", Integer.toString(tree.getType()));
			if (tree.hasLabel()) writer.writeAttribute("label", tree.getLabel());
		}

		@Override
		public void endTree(ITree tree) throws XMLStreamException {
			writer.writeEndElement();
		}
	}
}